import six
import logging
import copy

from uuid import uuid4
from collections import namedtuple, defaultdict
from altgraph.Graph import Graph

from rabix.common.errors import ValidationError, RabixError
from rabix.common.util import wrap_in_list
from rabix.common.models import App, IO, Job
from rabix.schema import JsonSchema

log = logging.getLogger(__name__)

AppNode = namedtuple('AppNode', ['app', 'inputs'])

Relation = namedtuple('Relation', ['src_port', 'dst_port'])
InputRelation = namedtuple('InputRelation', ['dst_port'])
OutputRelation = namedtuple('OutputRelation', ['src_port'])


class Step(object):

    def __init__(self, step_id, app, inputs, outputs=None):
        self.id = step_id
        self.app = app
        self.inputs = inputs
        self.outputs = outputs

    def to_dict(self, context):
        return {
            "id": self.id,
            "@type": "Step",
            "app": self.app.to_dict(context),
            "inputs": context.to_dict(self.inputs),
            "outputs": context.to_dict(self.outputs)
        }

    @classmethod
    def from_dict(cls, context, d):
        return cls(
            d["id"],
            context.from_dict(d['app']),
            context.from_dict(d.get('inputs')),
            context.from_dict(d.get('outputs')))


class WorkflowApp(App):

    def __init__(self, app_id, steps, context,
                 inputs=None, outputs=None,
                 app_description=None,
                 annotations=None,
                 platform_features=None):
        self.graph = Graph()
        self.inputs = inputs or []
        self.outputs = outputs or []
        self.executor = context.executor
        self.steps = steps

        for step in steps:
            self.add_node(step.id,  AppNode(step.app, {}))

        for step in steps:
            # inputs
            for input_port, input_val in six.iteritems(step.inputs):
                inp = wrap_in_list(input_val)
                for item in inp:
                    self.add_edge_or_input(step, input_port, item)

            # outputs
            if step.outputs:
                for output_port, output_val in six.iteritems(step.outputs):
                    output_node = step.app.get_output(output_port)
                    output_id = output_val['$to']
                    self.add_node(output_id, output_node)
                    self.graph.add_edge(
                        step.id, output_id, OutputRelation(output_port)
                    )
                    self.outputs.append(output_node)

        if not self.graph.connected():
            pass
            # raise ValidationError('Graph is not connected')

        schema = {
            "@type": "JsonSchema",
            "type": "object",
            "properties": {},
            "required": []
        }

        for inp in self.inputs:
            schema['properties'][inp.id] = inp.validator.schema
            if inp.required:
                schema['required'].append(inp.id)

        super(WorkflowApp, self).__init__(
            app_id, JsonSchema(context, schema), self.outputs,
            app_description=app_description,
            annotations=annotations,
            platform_features=platform_features
        )

    def add_edge_or_input(self, step, input_name, input_val):
        node_id = step.id
        if isinstance(input_val, dict) and '$from' in input_val:
            frm = wrap_in_list(input_val['$from'])
            for inp in frm:
                if '.' in inp:
                    node, outp = inp.split('.')
                    self.graph.add_edge(node, node_id, Relation(outp, input_name))
                else:
                    # TODO: merge input schemas if one input goes to different apps

                    input = step.app.get_input(input_name)
                    self.add_node(inp, input)
                    self.graph.add_edge(
                        inp, node_id, InputRelation(input_name)
                    )
                    wf_input = copy.deepcopy(input)
                    wf_input.id = inp
                    self.inputs.append(wf_input)

        else:
            self.graph.node_data(node_id).inputs[input_name] = input_val

    # Graph.add_node silently fails if node already exists
    def add_node(self, node_id, node):
        if node_id in self.graph.nodes:
            raise ValidationError('Duplicate node ID: %s' % node_id)
        self.graph.add_node(node_id, node)

    def hide_nodes(self, type):
        for node_id in self.graph.node_list():
            node = self.graph.node_data(node_id)
            if isinstance(node, type):
                self.graph.hide_node(node_id)

    def run(self, job):
        eg = ExecutionGraph(self, job)
        while eg.has_next():
            next_id, next = eg.next_job()
            self.executor.execute(next, eg.job_done, next_id)
        return eg.outputs

    def to_dict(self, context):
        d = super(WorkflowApp, self).to_dict(context)
        d.update({
            "@type": "Workflow",
            'steps': [step.to_dict(context) for step in self.steps]
        })
        return d

    @classmethod
    def from_dict(cls, context, d):
        steps = [Step(
            step['id'], context.from_dict(step['app']),
            step['inputs'], step.get('outputs')
        )
            for step in d['steps']]

        return cls(
            d.get('@id', str(uuid4())),
            steps,
            context
        )


def init(context):
    context.add_type('Workflow', WorkflowApp.from_dict)


class PartialJob(object):

    def __init__(self, node_id, app, inputs, input_counts, outputs):
        self.result = None
        self.status = 'WAITING'
        self.node_id = node_id
        self.app = app
        self.inputs = inputs
        self.input_counts = input_counts
        self.outputs = outputs

        self.running = []
        self.resources = None

    @property
    def resolved(self):
        for name, cnt in six.iteritems(self.input_counts):
            if cnt > 0:
                return False
        return True

    def resolve_input(self, input_port, results):
        log.debug("Resolving input '%s' with value %s" % (input_port, results))
        input_count = self.input_counts[input_port]
        if input_count <= 0:
            raise RabixError("Input already satisfied")
        self.input_counts[input_port] = input_count - 1
        # recursive_merge(self.job['inputs'].get(input_port), results)
        prev_result = self.inputs.get(input_port)
        if prev_result is None:
            self.inputs[input_port] = results
        elif isinstance(prev_result, list):
            prev_result.append(results)
        else:
            self.inputs[input_port] = [prev_result, results]
        return self.resolved

    def propagate_result(self, result):
        log.debug("Propagating result: %s" % result)
        self.result = result
        for k, v in six.iteritems(result):
            log.debug("Propagating result: %s, %s" % (k, v))
            self.outputs[k].resolve_input(v)

    def job(self):
        return Job(self.node_id, self.app, self.inputs, {})


class ExecRelation(object):

    def __init__(self, node, input_port):
        self.node = node
        self.input_port = input_port

    def resolve_input(self, result):
        self.node.resolve_input(self.input_port, result)


class OutRelation(object):

    def __init__(self, graph, name):
        self.name = name
        self.graph = graph

    def resolve_input(self, result):
        self.graph.outputs[self.name] = result


class ExecutionGraph(object):

    def __init__(self, workflow, job):
        self.workflow = workflow
        self.executables = {}
        self.ready = {}
        self.job = job
        self.outputs = {}

        graph = workflow.graph

        for node_id in graph.back_topo_sort()[1]:
            executable = self.make_executable(node_id)
            if executable:
                self.executables[node_id] = executable

        workflow.hide_nodes(IO)

        self.order = graph.back_topo_sort()[1]

    def make_executable(self, node_id):
        node = self.graph.node_data(node_id)
        if isinstance(node, IO):
            return None

        out_edges = self.graph.out_edges(node_id)
        in_edges = self.graph.inc_edges(node_id)

        outputs = {}
        input_counts = ExecutionGraph.count_inputs(self.graph, in_edges)
        for out_edge in out_edges:
            rel = self.graph.edge_data(out_edge)
            if isinstance(rel, Relation):
                tail = self.executables[self.graph.tail(out_edge)]
                outputs[rel.src_port] = ExecRelation(tail, rel.dst_port)
            elif isinstance(rel, OutputRelation):
                tail = self.graph.tail(out_edge)
                outputs[rel.src_port] = OutRelation(self, tail)

        executable = PartialJob(
            node_id, node.app, node.inputs, input_counts, outputs
        )

        for in_edge in in_edges:
            rel = self.graph.edge_data(in_edge)
            head = self.graph.head(in_edge)
            if (isinstance(rel, InputRelation) and
                    head in self.job.inputs):
                executable.resolve_input(
                    rel.dst_port, self.job.inputs[head]
                )

        return executable

    @staticmethod
    def count_inputs(graph, in_edges):
        input_counts = defaultdict(lambda: 0)
        for edge in in_edges:
            relation = graph.edge_data(edge)
            input_count = input_counts[relation.dst_port]
            input_counts[relation.dst_port] = input_count + 1
        return input_counts

    def job_done(self, node_id, results):
        ex = self.executables[node_id]
        ex.propagate_result(results)

    def next_job(self):
        if not self.order:
            return None
        next = self.order.pop()
        return next, self.executables[next].job()

    def has_next(self):
        return len(self.order) > 0

    @property
    def graph(self):
        return self.workflow.graph


# Smoke test
if __name__ == '__main__':
    from os.path import abspath, join
    from rabix.common.ref_resolver import from_url

    def root_relative(path):
        p = abspath(join(__file__, '../', path))
        return abspath(join(__file__, '../../../', path))

    doc = from_url(root_relative('tests/workflow.yml'))

    wf = WorkflowApp(doc['workflows']['add_one_mul_two']['steps'])
    print(wf.graph.forw_topo_sort())

    for edge in wf.graph.edges:
        print(wf.graph.describe_edge(edge))

    print(wf.inputs)
    print(wf.outputs)



from __future__ import print_function
import logging
import six
import shlex

from docker.errors import APIError
from docker.utils.utils import parse_repository_tag

from rabix.common.errors import ResourceUnavailable, RabixError

log = logging.getLogger(__name__)


def make_config(**kwargs):
    keys = ['Hostname', 'Domainname', 'User', 'Memory', 'MemorySwap',
            'CpuShares', 'Cpuset', 'AttachStdin', 'AttachStdout',
            'AttachStderr', 'PortSpecs', 'ExposedPorts', 'Tty', 'OpenStdin',
            'StdinOnce', 'Env', 'Cmd', 'Image', 'Volumes', 'WorkingDir',
            'Entrypoint', 'NetworkDisabled', 'OnBuild']
    cfg = {
        'AttachStdin': False,
        'AttachStdout': False,
        'AttachStderr': False,
        'Tty': False,
        'Privileged': False,
    }
    cfg.update({k[0].upper() + k[1:]: v for k, v in six.iteritems(kwargs)})
    cfg = {k: v for k, v in six.iteritems(cfg) if k in keys}
    entrypoint = cfg.get("Entrypoint")
    if isinstance(entrypoint, six.string_types):
        cfg['Entrypoint'] = shlex.split(entrypoint)
    return cfg


class Container(object):

    def __init__(self, docker_client, image_id, image_uri, cmd, user=None,
                 volumes=None, mem_limit=0, ports=None, environment=None,
                 entrypoint=None, cpu_shares=None, working_dir=None, **kwargs):
        self.docker_client = docker_client
        self.image_id = image_id
        self.uri = image_uri
        self.cmd = cmd
        self.user = user
        self.volumes = volumes
        self.mem_limit = mem_limit
        self.ports = ports
        self.environment = environment  # ["PASSWORD=xxx"]{"PASSWORD": "xxx"}
        self.entrypoint = entrypoint  #
        self.cpu_shares = cpu_shares
        self.working_dir = working_dir
        self.produced_image = None
        kwargs.update(
            {
                "Image": image_id,
                "Cmd": cmd,
                "User": user,
                "Volumes": volumes,
                "Memory": mem_limit,
                "ExposedPorts": ports,
                "Env": environment,
                "CpuShares": cpu_shares,
                "WorkingDir": working_dir
            })
        self.config = make_config(**kwargs)

        get_image(docker_client, image_id=self.image_id, repo=self.uri)
        self.container =\
            self.docker_client.create_container_from_config(self.config)

    def start(self, binds=None, port_bindings=None):
        try:
            self.docker_client.start(container=self.container, binds=binds,
                                     port_bindings=port_bindings)
        except APIError:
            logging.error('Failed to run container %s' % self.container)
            raise RabixError('Unable to run container from image %s:'
                             % self.image_id)

    def remove(self, success_only=False):
        self.wait()
        if not success_only or self.is_success():
            self.docker_client.remove_container(self.container)
        return self

    def inspect(self):
        return self.docker_client.inspect_container(self.container)

    def is_running(self):
        return self.inspect()['State']['Running']

    def wait(self):
        if self.is_running():
            self.docker_client.wait(self.container)
        return self

    def is_success(self):
        return self.wait().inspect()['State']['ExitCode'] == 0

    def write_stdout(self, file=None):
        write = lambda out: print(out.rstrip())
        f = None
        if file:
            f = open(file, 'w', buffering=1)
            write = f.write

        if self.is_running():
            for out in self.docker_client.attach(self.container, stdout=True,
                                                 stderr=False, stream=True,
                                                 logs=True):
                write(out)
            if f:
                f.close()
        else:
            print(self.docker_client.logs(self.container))
        return self

    def commit(self, message=None, conf=None, repository=None, tag=None):
        log.debug("repository: {}, ")
        self.produced_image = self.docker_client.commit(
            self.container['Id'], message=message, conf=make_config(**conf),
            repository=repository, tag=tag
        )
        return self


def match_image(image, query):
    """Returns image dict if it exists locally, or None"""
    if isinstance(query, str):
        return (len(query) >= 12 and image['Id'].startswith(query)) or \
            query in image['RepoTags'] or \
            (query + ":latest") in image['RepoTags']

    if isinstance(query, list):
        return any([match_image(image, q) for q in query])

    if isinstance(query, tuple):
        repo, tag = query
        return (repo + ":" + tag) in image['RepoTags']

    if isinstance(query, dict):
        image_id = query.get('image_id')
        repo = query.get('repo')
        tag = query.get('tag', 'latest')
        return (image_id and image['Id'].startswith(image_id)) or \
               ((repo + ":" + tag) in image['RepoTags'])

    return False


def find_image(client, query):
    images = client.images()
    it = (image for image in images if match_image(image, query))
    return next(it, None)


def get_image(client, repo=None, tag=None, image_id=None):
    """Returns the image dict. Pulls from repo if not found locally."""

    if not image_id and not repo:
        raise ValueError('Need either repository or image ID.')

    if repo and not tag:
        repo, tag = parse_repository_tag(repo)

    queries = [image_id, repo]
    if tag:
        queries.append((repo, tag))

    img = find_image(client, queries)

    if not img:
        log.info('Pulling %s', repo)
        client.pull(repo, tag)
        img = find_image(client, queries)

    if not img:
        raise ResourceUnavailable(repo, 'Image not found.')

    return img



import docopt
import logging
import yaml

from os.path import isfile

from rabix import __version__ as version
from rabix.common.errors import RabixError
from rabix.common.ref_resolver import Loader
from rabix.common.util import log_level
from rabix.tools.build import run_steps


log = logging.getLogger(__name__)


USAGE = """
Usage:
  rabix-tools build [-v...] [--config=<cfg_path>]
  rabix-tools checksum [--method=(md5|sha1)] <jsonptr>
  rabix-tools -h | --help
  rabix-tools --version

Commands:
  build                     Execute steps for app building, wrapping and
                            deployment.
  checksum                  Calculate and print the checksum of json document
                            (or fragment) pointed by <jsonptr>

Options:
  -c --config=<cfg_path>    Specify path to config file [default: .rabix.yml]
  -h --help                 Display this message.
  -m --method (md5|sha1)    Checksum type [default: sha1]
  --version                 Print version to standard output and quit.
  -v --verbose              Verbosity. More Vs more output.
"""


def checksum(jsonptr, method='sha1'):
    loader = Loader()
    obj = loader.load(jsonptr)
    print(method + '$' + loader.checksum(obj, method))


def build(path='.rabix.yml'):
    if not isfile(path):
        raise RabixError('Config file %s not found!' % path)
    with open(path) as cfg:
        config = yaml.load(cfg)
        run_steps(config)


def main():
    logging.basicConfig(level=logging.WARN)

    args = docopt.docopt(USAGE, version=version)

    logging.root.setLevel(log_level(args['--verbose']))
    if args["checksum"]:
        checksum(args['<jsonptr>'], args['--method'])
    elif args["build"]:
        build(args.get("--config"))


if __name__ == '__main__':
    main()



import six
from rabix.common.models import IO
from rabix.common.errors import RabixError


class JsonSchema(object):

    def __init__(self, context, schema):

        if 'type' not in schema:
            raise RabixError(
                "Invalid JSON schema: schema doesn't have a type.")

        if schema['type'] != 'object':
            raise RabixError("Invalid JSON schema: schema type isn't 'object'")

        if 'properties' not in schema:
            raise RabixError(
                "Invalid JSON schema: schema doesn't have properties")

        self.schema = schema
        self.schema['@type'] = 'JsonSchema'
        required = schema.get('required', [])
        self.io = [
            IO.from_dict(context, {
                '@id': k,
                'schema': v,
                'required': k in required,
                'annotations': v.get('adapter')
            })
            for k, v in six.iteritems(schema['properties'])
        ]

    def __iter__(self, *args, **kwargs):
        return self.io.__iter__(*args, **kwargs)

    def to_dict(self, context=None):
        return self.schema


def init(context):
    context.add_type('JsonSchema', JsonSchema)



from rabix.main import get_tool, dry_run_parse


def test_cmd_line():
    cmd1 = dry_run_parse(['https://s3.amazonaws.com/rabix/rabix-test/'
                          'bwa-mem.json',
                          '-i', './rabix/tests/test-cmdline/inputs.json'])
    tool1 = get_tool(cmd1)
    assert tool1



import nose
import six

from nose.tools import *

from rabix.schema import JsonSchema
from rabix.main import init_context
from rabix.common.errors import RabixError


def test_simple_json_schema():
    schema = {
        "type": "object",
        "properties": {
            "a": {
                "type": "integer"
            }
        }
    }

    js = JsonSchema(init_context(), schema)
    a = next(iter(js))
    assert_equal(a.id, 'a')
    assert_equal(a.constructor, int)
    assert_false(a.required)

    to_dict = js.to_dict()
    for k, v in six.iteritems(schema):
        assert_equal(to_dict[k], v)


def test_invalid_json_schema():
    ctx = init_context()

    no_type = {
        "properties": {
            "a": {
                "type": "integer"
            }
        }
    }
    assert_raises(RabixError, JsonSchema, ctx, no_type)

    not_object = {
        "type": "int",
        "properties": {
            "a": {
                "type": "integer"
            }
        }
    }
    assert_raises(RabixError, JsonSchema, ctx, not_object)

    no_properties = {
        "type": "object",
    }
    assert_raises(RabixError, JsonSchema, ctx, no_properties)


if __name__ == '__main__':
    nose.run()



import six

from rabix.common.errors import ValidationError


class Context(object):

    def __init__(self, type_map, executor):
        self.type_map = type_map
        self.executor = executor

    def add_type(self, name, constructor):
        self.type_map[name] = constructor

    def from_dict(self, d):
        if d is None:
            return None
        if isinstance(d, list):
            return [self.from_dict(e) for e in d]
        if not isinstance(d, dict):
            return d
        if '@type' not in d:
            return {k: self.from_dict(v) for k, v in six.iteritems(d)}
        type_name = d['@type']
        constructor = self.type_map.get(type_name)
        if not constructor:
            raise ValidationError("Unknown type: %s" % type_name)
        return constructor(self, d)

    def to_dict(self, o):
        if o is None:
            return None
        if isinstance(o, dict):
            return {k: self.to_dict(v) for k, v in six.iteritems(o)}
        if isinstance(o, list):
            return [self.to_dict(e) for e in o]
        if hasattr(o, 'to_dict'):
            return o.to_dict(self)
        if isinstance(o, (int, float, bool, six.string_types)):
            return o
        raise RuntimeError("can't transform %s to dict" % o)



import random
import itertools
import collections
import logging
import six

from rabix.common.errors import RabixError

log = logging.getLogger(__name__)


def wrap_in_list(val, *append_these):
    """
    >>> wrap_in_list(1, 2)
    [1, 2]
    >>> wrap_in_list([1, 2], 3, 4)
    [1, 2, 3, 4]
    >>> wrap_in_list([1, 2], [3, 4])
    [1, 2, [3, 4]]
    """
    wrapped = val if isinstance(val, list) else [val]
    return wrapped + list(append_these)


def import_name(name):
    name = str(name)
    if '.' not in name:
        return __import__(name)
    chunks = name.split('.')
    var_name = chunks[-1]
    module_name = '.'.join(chunks[:-1])
    fromlist = chunks[:-2] if len(chunks) > 2 else []
    module = __import__(module_name, fromlist=fromlist)
    if not hasattr(module, var_name):
        raise ImportError('%s not found in %s' % (var_name, module_name))
    return getattr(module, var_name)


def dot_update_dict(dst, src):
    for key, val in six.iteritems(src):
        t = dst
        if '.' in key:
            for k in key.split('.'):
                if k == key.split('.')[-1]:
                    if isinstance(val, collections.Mapping):
                        t = t.setdefault(k, {})
                        dot_update_dict(t, src[key])
                    else:
                        t[k] = val
                else:
                    if not isinstance(t.get(k), collections.Mapping):
                        t[k] = {}
                    t = t.setdefault(k, {})
        else:
            if isinstance(val, collections.Mapping):
                t = t.setdefault(key, {})
                dot_update_dict(t, src[key])
            else:
                t[key] = val
    return dst


def rnd_name(syllables=5):
    return ''.join(itertools.chain(*zip(
        (random.choice('bcdfghjklmnpqrstvwxz') for _ in range(syllables)),
        (random.choice('aeiouy') for _ in range(syllables)))))


def log_level(int_level):
    if int_level <= 0:
        level = logging.WARN
    elif int_level == 1:
        level = logging.INFO
    elif int_level >= 2:
        level = logging.DEBUG
    else:
        raise RabixError("Invalid log level: %s" % int_level)
    return level


def sec_files_naming_conv(path, ext):
    return ''.join([path, ext]) if ext.startswith('.') \
        else '.'.join(['.'.join(path.split('.')[:-1]), ext])



import json
import logging

import execjs

from rabix.expressions import evaluator

log = logging.getLogger(__name__)


class JSEval(evaluator.ExpressionEvalPlugin):

    def __init__(self):
        super(JSEval, self).__init__()

    def evaluate(self, expression=None, job=None, context=None, *args,
                 **kwargs):
        log.debug("expression: %s" % expression)
        if expression.startswith('{'):
            exp_tpl = '''function () {
            $job = %s;
            $self = %s;
            return function()%s();}()
            '''
        else:
            exp_tpl = '''function () {
            $job = %s;
            $self = %s;
            return %s;}()
            '''
        exp = exp_tpl % (json.dumps(job), json.dumps(context), expression)
        log.debug("exec code: %s" % exp)

        result = execjs.eval(exp)
        log.debug("result: %s" % result)
        return result



import os
import six
import json
import logging
import time
import datetime

from uuid import uuid4
from jsonschema.validators import Draft4Validator
from six.moves.urllib.parse import urlparse


log = logging.getLogger(__name__)


class App(object):

    def __init__(self, app_id, inputs, outputs, app_description=None,
                 annotations=None, platform_features=None):
        self.id = app_id
        self.inputs = inputs
        self.outputs = outputs
        self.app_description = app_description
        self.annotations = annotations
        self.platform_features = platform_features
        self._inputs = {io.id: io for io in inputs}
        self._outputs = {io.id: io for io in outputs}

    def install(self, *args, **kwargs):
        pass

    def run(self, job):
        raise NotImplementedError("Method 'run' is not implemented"
                                  " in the App class")

    def get_input(self, name):
        return self._inputs.get(name)

    def get_output(self, name):
        return self._outputs.get(name)

    def validate_inputs(self, input_values):
        for inp in self.inputs:
            if inp.id in input_values:
                if not inp.validate(input_values[inp.id]):
                    return False
            elif inp.required:
                return False
        return True

    def job_dump(self, job, dirname, context):
        with open(os.path.join(dirname, 'job.cwl.json'), 'w') as f:
            json.dump(job.to_dict(context), f)
            log.info('File %s created.', job.id)

    def to_dict(self, context):
        return {
            '@id': self.id,
            '@type': 'App',
            'inputs': [context.to_dict(inp) for inp in self.inputs],
            'outputs': [context.to_dict(outp) for outp in self.outputs],
            'appDescription': self.app_description,
            'annotations': self.annotations,
            'platformFeatures': self.platform_features
        }


class File(object):

    def __init__(self, path, size=None, meta=None, secondary_files=None):
        self.url = path
        self.size = size
        self.meta = meta or {}
        self.secondary_files = secondary_files or []

    def to_dict(self, context):
        if isinstance(self.url, str):
            path = self.url
        else:
            path = self.url.geturl()
        return {
            "@type": "File",
            "path": path,
            "size": self.size,
            "metadata": self.meta,
            "secondaryFiles": [sf.to_dict(context)
                               for sf in self.secondary_files]
        }

    @classmethod
    def from_dict(cls, d):

        if isinstance(d, six.string_types):
            d = {'path': d}

        size = d.get('size')
        if size is not None:
            size = int(size)

        return cls(path=d.get('path'),
                   size=size,
                   meta=d.get('meta'),
                   secondary_files=[File.from_dict(sf)
                                    for sf in d.get('secondaryFiles', [])])


class IO(object):

    def __init__(self, port_id, validator=None, constructor=None,
                 required=False, annotations=None, depth=0):
        self.id = port_id
        self.validator = Draft4Validator(validator)
        self.required = required
        self.annotations = annotations
        self.constructor = constructor or str
        self.depth = depth

    def validate(self, value):
        return self.validator.validate(value)

    def to_dict(self, ctx=None):
        return {
            '@id': self.id,
            '@type': 'IO',
            'depth': self.depth,
            'schema': self.validator.schema,
            'required': self.required,
            'annotations': self.annotations
        }

    @classmethod
    def from_dict(cls, context, d):
        constructor_map = {
            'integer': int,
            'number': float,
            'boolean': bool,
            'object': dict,
            'string': str,
            'file': File.from_dict,
            'directory': File.from_dict
        }
        item_schema = d.get('schema', {})
        type_name = item_schema.get('type')
        depth = 0
        while type_name == 'array':
            depth += 1
            item_schema = item_schema.get('items', {})
            type_name = item_schema.get('type')

        return cls(d.get('@id', str(uuid4())),
                   validator=context.from_dict(d.get('schema')),
                   constructor=constructor_map[type_name],
                   required=d['required'],
                   annotations=d['annotations'],
                   depth=depth)


class Job(object):

    def __init__(self, job_id, app, inputs, allocated_resources):
        # if not app.validate_inputs(inputs):
        #     raise ValidationError("Invalid inputs for application %s" % app.id)
        self.id = job_id
        self.app = app
        self.inputs = inputs
        self.allocated_resources = allocated_resources

    def run(self):
        return self.app.run(self)

    def to_dict(self, context):
        return {
            '@id': self.id,
            '@type': 'Job',
            'app': self.app.to_dict(context),
            'inputs': context.to_dict(self.inputs),
            'allocatedResources': self.allocated_resources
        }

    @staticmethod
    def mk_work_dir(name):
        ts = time.time()
        path = '_'.join([name, datetime.datetime.fromtimestamp(ts).strftime('%H%M%S')])
        try_path = path
        if os.path.exists(path):
            num = 0
            try_path = path
            while os.path.exists(try_path):
                try_path = '_'.join([path, str(num)])
                num += 1
        return try_path

    @classmethod
    def from_dict(cls, context, d):
        app = context.from_dict(d['app'])
        return cls(
            d.get('@id') if d.get('@id') else cls.mk_work_dir(app.id), app,
            context.from_dict(d['inputs']), d.get('allocatedResources')
        )


class Resource(object):
    pass



import six

from uuid import uuid4

from rabix.common.errors import RabixError
from rabix.common.models import App
from rabix.expressions.evaluator import Evaluator


class ScriptApp(App):

    def __init__(self, app_id, inputs, outputs, script, context,
                 app_description=None,
                 annotations=None,
                 platform_features=None):
        super(ScriptApp, self).__init__(
            app_id, inputs, outputs,
            app_description=app_description,
            annotations=annotations,
            platform_features=platform_features
        )
        self.script = script
        self.evaluator = Evaluator()
        self.context = context

    def run(self, job):
        if isinstance(self.script, six.string_types):
            lang = 'javascript'
            expr = self.script
        elif isinstance(self.script, dict):
            lang = self.script['lang']
            expr = self.script['value']
        else:
            raise RabixError("invalid script")

        result = self.evaluator.evaluate(lang, expr, job.to_dict(self.context), None)
        return result

    def to_dict(self, context):
        d = super(ScriptApp, self).to_dict(context)
        d.update({
            "@type": "Script",
            "script": self.script,
            "inputs": self.inputs.schema,
            "outputs": self.outputs.schema
        })
        return d

    @classmethod
    def from_dict(cls, context, d):
        return cls(d.get('@id', str(uuid4())),
                   context.from_dict(d['inputs']),
                   context.from_dict(d['outputs']),
                   d['script'],
                   context,
                   app_description=d.get('appDescription'),
                   annotations=d.get('annotations'),
                   platform_features=d.get('platform_features'))



import six


class RabixError(BaseException):
    def __init__(self, message=''):
        super(RabixError, self).__init__(message)
        self.message = message


class ValidationError(RabixError):
    def __init__(self, message):
        super(ValidationError, self).__init__(message)


class ResourceUnavailable(RabixError):
    def __init__(self, uri, message='', cause=None):
        msg = 'Unable to load "%s".' % uri
        if message:
            msg = ' '.join([msg, message])
        if cause:
            msg = ' '.join([msg, 'Reason:', six.text_type(cause)])
        super(ResourceUnavailable, self).__init__(msg)
        self.__cause__ = cause
        self.uri = uri



import mock

from nose.tools import raises, eq_
from os.path import abspath

import rabix.tools.build as steps
from rabix.common.errors import RabixError


@mock.patch('rabix.tools.build.get_image')
@mock.patch('rabix.tools.build.Container')
def test_run_ok(container_mock, get_image_mock):
    container_mock().is_success = mock.Mock(return_value=True)

    steps.run('docker_client', 'image_id', cmd=['cmd'])

    assert_run(container_mock)


@raises(RabixError)
@mock.patch('rabix.tools.build.get_image')
@mock.patch('rabix.tools.build.Container')
def test_run_fail(container_mock, get_image_mock):
    container_mock().is_success = mock.Mock(return_value=False)

    try:
        steps.run('docker_client', 'image_id',
                  cmd=['cmd'], mount_point=steps.MOUNT_POINT)
    finally:
        assert_run(container_mock)


def assert_run(container):
    # container.assert_called_with('docker_client', 'image_id',
    #                             {}, mount_point=steps.MOUNT_POINT)
    instance = container()
    instance.start.assert_called_with({abspath('.'): steps.MOUNT_POINT})
    instance.is_success.assert_called_with()


@mock.patch('rabix.tools.build.get_image')
@mock.patch('rabix.tools.build.Container')
def test_build_ok(container_mock, get_image_mock):
    c = container_mock()
    c.is_success = mock.Mock(return_value=True)

    steps.build('docker_client', 'image_id', cmd=['cmd'], message='message',
                register={'repo': 'repo', 'tag': 'tag'})

    container_mock.assert_called_with('docker_client', 'image_id',
                                      {},
                                      mount_point=steps.MOUNT_POINT)
    c.run.assert_called_with(['cmd'])
    c.is_success.assert_called_with()
    c.commit.assert_called_with('message', {'Cmd': []},
                                tag='tag', repository='repo')


@mock.patch('rabix.tools.build.get_image')
@mock.patch('rabix.tools.build.Container')
def test_build_ok(container_mock, get_image_mock):
    container_mock().is_success = mock.Mock(return_value=True)

    steps.build('docker_client', 'image_id', cmd=['cmd'], message='message',
                register={'repo': 'repo', 'tag': 'tag'})

    assert_build(container_mock)
    container_mock().commit.assert_called_with('message', {'Cmd': []},
                                               tag='tag', repository='repo')


@raises(RabixError)
@mock.patch('rabix.tools.build.get_image')
@mock.patch('rabix.tools.build.Container')
def test_build_fail(container_mock, get_image_mock):
    container_mock().is_success = mock.Mock(return_value=False)

    try:
        steps.build('docker_client', 'image_id', cmd=['cmd'], message='message',
                    register={'repo': 'repo', 'tag': 'tag'})
    finally:
        assert_build(container_mock)
        assert not container_mock().commit.called


def assert_build(container):
    # container.assert_called_with('docker_client', 'image_id',
    #                              {},
    #                              mount_point=steps.MOUNT_POINT)
    instance = container()
    instance.start.assert_called_with({abspath('.'): steps.MOUNT_POINT})
    instance.is_success.assert_called_with()


def test_make_cmd():
    eq_(steps.make_cmd('cmd'), ['cmd'])
    eq_(steps.make_cmd('cmd --param "file name"'),
        ['cmd', '--param', 'file name'])
    eq_(steps.make_cmd(['cmd']), ['cmd'])
    eq_(steps.make_cmd(['cmd', '--param']), ['cmd', '--param'])
    eq_(steps.make_cmd(['cmd'], join=True), ['cmd'])
    eq_(steps.make_cmd(['cmd'], join=True), ['cmd'])
    eq_(steps.make_cmd(['cmd1 --param1', 'cmd2 --param2'], join=True),
        ['/bin/sh', '-c', 'cmd1 --param1 && cmd2 --param2'])


def test_resolve():
    r = steps.Runner(None, context={"x": "5"})
    eq_(r.resolve("something"), "something")
    eq_(r.resolve("${x}"), "5")
    eq_(r.resolve(["${x}", "y"]), ["5", "y"])
    eq_(r.resolve({"${y}": "${x}"}), {"${y}": "5"})



import rabix
import rabix.common

from nose.tools import *
from rabix.common.util import *


def test_wrap():
    assert_equal(wrap_in_list(1, 2), [1, 2])
    assert_equal(wrap_in_list([], 1, 2), [1, 2])
    assert_equal(wrap_in_list([1, 2], 3, 4), [1, 2, 3, 4])
    assert_equal(wrap_in_list([1, 2], [3, 4]), [1, 2, [3, 4]])


def test_import_name():
    m = import_name('rabix')
    assert_equal(m, rabix)

    m = import_name('rabix.common')
    assert_equal(m, rabix.common)

    assert_raises(ImportError, import_name, 'rabix.badmodule')


def test_dot_update_dict():
    assert_equal(dot_update_dict({}, {}), {})
    assert_equal(dot_update_dict({"a": 5}, {}), {"a": 5})
    assert_equal(dot_update_dict({}, {"a": 5}), {"a": 5})
    assert_equal(dot_update_dict({"b": 4}, {"a": 5}), {"a": 5, "b": 4})
    assert_equal(dot_update_dict({"a": 4}, {"a": 5}), {"a": 5})
    assert_equal(dot_update_dict({"a": 4}, {"a": 5}), {"a": 5})

    assert_equal(dot_update_dict({"a": 4}, {"b.c": 5}),
                 {"a": 4, "b": {"c": 5}})

    assert_equal(dot_update_dict({"b": {"a": 4}}, {"b.c": 5}),
                 {"b": {"c": 5, "a": 4}})

    assert_equal(dot_update_dict({"b": {"a": 4}}, {"b": {"c": 5}}),
                 {"b": {"c": 5, "a": 4}})

    assert_equal(dot_update_dict({"b": {"a": 4}}, {"b.c": {"d": 5}}),
                 {"b": {"c": {"d": 5}, "a": 4}})


def test_rnd_name():
    syllables = 6
    name = rnd_name(syllables)
    assert isinstance(name, six.string_types)
    assert_equal(len(name), syllables * 2)


def test_log_level():
    assert_equals(log_level(-5), logging.WARN)
    assert_equals(log_level(0), logging.WARN)
    assert_equals(log_level(1), logging.INFO)
    assert_equals(log_level(2), logging.DEBUG)
    assert_equals(log_level(99), logging.DEBUG)
    assert_raises(RabixError, log_level, 0.5)


def test_sec_files_naming_conv():
    assert_equals(sec_files_naming_conv("bla.ext", '.a'), "bla.ext.a")
    assert_equals(sec_files_naming_conv("bla.ext", 'b'), "bla.b")



import copy
import operator
import six
import logging
import os
import glob

from jsonschema import Draft4Validator
import jsonschema.exceptions

from six.moves import reduce
from rabix.common.ref_resolver import resolve_pointer
from rabix.common.util import sec_files_naming_conv
from rabix.expressions import Evaluator

ev = Evaluator()
log = logging.getLogger(__name__)


def evaluate(expr_object, job, context=None, *args, **kwargs):
    if isinstance(expr_object, dict):
        return ev.evaluate(expr_object.get('lang', 'javascript'), expr_object.get(
            'value'), job, context, *args, **kwargs)
    else:
        return ev.evaluate('javascript', expr_object, job, context, *args, **kwargs)


def intersect_dicts(d1, d2):
    return {k: v for k, v in six.iteritems(d1) if v == d2.get(k)}


def eval_resolve(val, job, context=None):
    if not isinstance(val, dict):
        return val
    if 'expr' in val or '$expr' in val:
        v = val.get('expr') or val.get('$expr')
        return evaluate(v, job, context)
    if 'job' in val:
        return resolve_pointer(job, val['job'], default=None)
    return val


class InputAdapter(object):
    def __init__(self, value, job_dict, schema, adapter_dict=None, key=''):
        self.job = job_dict
        self.schema = schema or {}
        self.adapter = adapter_dict or self.schema.get('adapter')
        self.has_adapter = self.adapter is not None
        self.adapter = self.adapter or {}
        self.key = key
        if 'oneOf' in self.schema:
            for opt in self.schema['oneOf']:
                validator = Draft4Validator(opt)
                try:
                    validator.validate(value)
                    self.schema = opt
                except jsonschema.exceptions.ValidationError:
                    pass
        self.value = eval_resolve(value, self.job)
        # if self.transform:
        #     self.value = eval_resolve(self.transform, self.job, self.value)
        if self.is_file():
            self.value = self.value['path']

    __str__ = lambda self: str(self.value)
    __repr__ = lambda self: 'InputAdapter(%s)' % self
    position = property(lambda self: (self.adapter.get('order', 9999999), self.key))
    # transform = property(lambda self: self.adapter.get('value'))
    prefix = property(lambda self: self.adapter.get('prefix'))
    item_separator = property(lambda self: self.adapter.get('itemSeparator', ','))

    @property
    def separator(self):
        return self.adapter.get('separator', ' ')

    def is_file(self):
        return isinstance(self.value, dict) and 'path' in self.value

    def arg_list(self):
        if isinstance(self.value, dict):
            return self.as_dict()
        if isinstance(self.value, list):
            return self.as_list()
        return self.as_primitive()

    def as_primitive(self):
        if self.value in (None, False):
            return []
        if self.value is True:
            if not self.prefix:
                raise ValueError('Boolean arguments must have a prefix.')
            return [self.prefix]
        if not self.prefix:
            return [self.value]
        if self.separator in [' ', None]:
            return [self.prefix, self.value]
        return [self.prefix + self.separator + six.text_type(self.value)]

    def as_dict(self, mix_with=None):
        sch = lambda key: self.schema.get('properties', {}).get(key, {})
        adapters = [InputAdapter(v, self.job, sch(k), key=k)
                    for k, v in six.iteritems(self.value)]
        adapters = (mix_with or []) + [adp for adp in adapters if adp.has_adapter]
        return reduce(
            operator.add,
            [a.arg_list() for a in sorted(adapters, key=lambda x: x.position)],
            []
        )

    def as_list(self):
        items = [InputAdapter(item, self.job, self.schema.get('items', {}))
                 for item in self.value]
        if not self.prefix:
            return reduce(operator.add, [a.arg_list() for a in items], [])
        if self.separator is None and self.item_separator is None:
            return reduce(operator.add, [[self.prefix] + a.arg_list() for a in items], [])
        if self.separator is not None and self.item_separator is None:
            return [self.prefix + self.separator + a.list_item() for a in items if a.list_item() is not None]
        joined = self.item_separator.join(filter(None, [a.list_item() for a in items]))
        if self.separator is None and self.item_separator is not None:
            return [self.prefix, joined]
        return [self.prefix + self.separator + joined]

    def list_item(self):
        as_arg_list = self.arg_list()
        if not as_arg_list:
            return None
        if len(as_arg_list) > 1 or isinstance(as_arg_list[0], (dict, list)):
            raise ValueError('Only lists of primitive values can use itemSeparator.')
        return six.text_type(as_arg_list[0])


class CLIJob(object):
    def __init__(self, job_dict, app, path_mapper=lambda x: x):
        self.job = copy.deepcopy(job_dict)
        self.app = app
        self.path_mapper = path_mapper
        self.rewrite_paths(self.job['inputs'])
        self.adapter = self.app.adapter or {}
        self.stdin = eval_resolve(self.adapter.get('stdin'), self.job)
        self.stdout = eval_resolve(self.adapter.get('stdout'), self.job)
        self.base_cmd = self.adapter.get('baseCmd', [])
        if isinstance(self.base_cmd, six.string_types):
            self.base_cmd = self.base_cmd.split(' ')
        self.args = self.adapter.get('args', [])
        self.input_schema = self.app.inputs.schema
        self.output_schema = self.app.outputs.schema

    def make_arg_list(self):
        adapters = [InputAdapter(a['value'], self.job, {}, a) for a in self.args]
        args = InputAdapter(self.job['inputs'], self.job, self.input_schema).as_dict(adapters)
        base_cmd = [eval_resolve(item, self.job) for item in self.base_cmd]
        return [six.text_type(arg) for arg in base_cmd + args]

    def cmd_line(self):
        a = self.make_arg_list()
        if self.stdin:
            a += ['<', self.stdin]
        if self.stdout:
            a += ['>', self.stdout]
        return ' '.join(a)  # TODO: escape

    def rewrite_paths(self, val):
        if isinstance(val, list):
            for item in val:
                self.rewrite_paths(item)
        elif isinstance(val, dict) and 'path' in val:
            val['path'] = self.path_mapper(val['path'])
        elif isinstance(val, dict):
            for item in six.itervalues(val):
                self.rewrite_paths(item)

    def get_outputs(self, job_dir):
        result, outs = {}, self.output_schema.get('properties', {})
        for k, v in six.iteritems(outs):
            adapter = v['adapter']

            files = glob.glob(os.path.join(job_dir, eval_resolve(adapter['glob'], self.job)))
            result[k] = [{'path': p, 'metadata': self._meta(p, adapter),
                          'secondaryFiles': self._secondaryFiles(p, adapter)} for p in files]
            if v['type'] != 'array':
                result[k] = result[k][0] if result[k] else None
        return result

    def _meta(self, path, adapter):
        meta, result = adapter.get('metadata', {}), {}
        inherit = meta.pop('__inherit__', None)
        if inherit:
            src = self.job['inputs'].get(inherit)
            if isinstance(src, list):
                result = reduce(intersect_dicts, [x.get('metadata', {}) for x in src]) \
                    if len(src) > 1 else src[0].get('metadata', {})
            elif isinstance(src, dict):
                result = src.get('metadata', {})
        result.update(**meta)
        for k, v in six.iteritems(result):
            result[k] = eval_resolve(v, self.job, context=path)
        return result

    def _secondaryFiles(self, p, adapter):
        secondaryFiles = []
        secFiles = adapter.get('secondaryFiles', [])
        for s in secFiles:
            path = sec_files_naming_conv(p, s)
            secondaryFiles.append({'path': path})
        return secondaryFiles



import logging
import re
import six
import shlex

from os.path import abspath
from docker.utils.utils import parse_repository_tag

from rabix.docker.container import Container, get_image
from rabix.docker import docker_client
from rabix.common.errors import RabixError

log = logging.getLogger(__name__)

MOUNT_POINT = '/build'


def build(client, from_img, **kwargs):
    container = run_container(client, from_img, kwargs, {})
    if container.is_success():
        message = kwargs.pop('message', None)
        register = kwargs.pop('register', {})
        cfg = {"Cmd": []}
        cfg.update(**kwargs)

        container.commit(
            message, cfg, repository=register.get('repo'),
            tag=register.get('tag')
        )
    else:
        raise RabixError("Build failed!")
    return container.produced_image['Id']


def run(client, from_img, **kwargs):
    container = run_container(client, from_img, kwargs, kwargs)
    if not container.is_success():
        raise RabixError(container.docker_client.logs(container.container))


def run_container(client, from_img, kwargs, container_kwargs):

    cmd = kwargs.pop('cmd', None)
    if not cmd:
        raise RabixError("Commands ('cmd') not specified!")

    repo, tag = parse_repository_tag(from_img)
    img = get_image(client, from_img)
    if not img:
        raise RabixError("Unable to find image: %s" % img)

    mount_point = kwargs.pop('mount_point', MOUNT_POINT)
    run_cmd = make_cmd(cmd, join=True)

    container = Container(client, img['Id'],
                          "docker://{}#{}".format(repo, tag),
                          run_cmd, volumes={mount_point: {}},
                          working_dir=mount_point, **container_kwargs)

    container.start({abspath('.'): mount_point})
    container.write_stdout()
    return container


def make_cmd(cmd, join=False):
    if isinstance(cmd, six.string_types):
        return shlex.split(cmd)
    elif isinstance(cmd, list) and len(cmd) > 1 and join:
        return ['/bin/sh', '-c', ' && '.join(cmd)]
    return cmd


class Runner(object):

    def __init__(self, docker, steps=None, context=None):
        self.types = {
            "run": run,
            "build": build
        }
        self.types.update(steps or {})

        self.context = context or {}

        self.docker = docker
        pass

    def run(self, config):
        steps = config['steps']
        for step in steps:
            step_name, step_conf = step.popitem()
            type_name = step_conf.pop('type', None)
            if not type_name:
                raise RabixError("Step type not specified!")

            step_type = self.types.get(type_name)
            if not step_type:
                raise RabixError("Unknown step type: %s" % type_name)

            resolved = {k: self.resolve(v) for k, v
                        in six.iteritems(step_conf)}

            img = resolved.pop('from', None)
            if not img:
                raise RabixError("Base image ('from') not specified!")

            log.info("Running step: %s" % step_name)
            self.context[step_name] = \
                step_type(self.docker, img, **resolved)
        pass

    def resolve(self, val):
        if isinstance(val, list):
            return [self.resolve(item) for item in val]
        elif isinstance(val, dict):
            return {k: self.resolve(v) for k, v in six.iteritems(val)}
        elif isinstance(val, six.string_types):
            resolved = re.sub("\$\{([a-zA-Z0-9_]+)\}",
                              lambda x: self.context[x.group(1)],
                              val)

            return resolved
        else:
            return val


def run_steps(config, docker_host=None, steps=None, context=None):
    r = Runner(docker_client(docker_host), steps, context)
    r.run(config)



import os
import tempfile
import logging
import uuid
import json
import six
import requests
import glob

from six.moves.urllib.parse import urlparse, urlunparse
from six.moves import input as raw_input
from rabix.common.errors import ResourceUnavailable
from rabix.common.util import sec_files_naming_conv
from rabix.common.ref_resolver import from_url


log = logging.getLogger(__name__)


def to_json(obj, fp=None):
    default = lambda o: (o.__json__() if callable(getattr(o, '__json__',
                                                          None))
                         else six.text_type(o))
    kwargs = dict(default=default, indent=2, sort_keys=True)
    return json.dump(obj, fp, **kwargs) if fp else json.dumps(obj, **kwargs)


class InputCollector(object):
    """
    Will handle local files, 'data:,' URLs (for tests) and delegate other
    URLs to requests.get()
    """
    def __init__(self):
        self.downloader = self.detect_downloader()

    @staticmethod
    def detect_downloader():
        pass

    @property
    def task_dir(self):
        if not self.dir:
            self.dir = str(uuid.uuid4())
        if not os.path.exists(self.dir):
            os.mkdir(self.dir)
        return self.dir

    def set_dir(self, job_dir):
        self.dir = job_dir

    def download(self, url, secondaryFiles=None, prompt=True):
        npath = self._download(url, metasearch=True)
        if os.path.exists(npath + '.rbx.json'):
            file = from_url(url + '.rbx.json')
            startdir = os.path.dirname(url)
            for i, v in enumerate(file.get('secondaryFiles', [])):
                spath = v['path']
                if not os.path.isabs(spath) and urlparse(spath, 'file').scheme == 'file':
                    spath = os.path.join(startdir, spath)
                file['secondaryFiles'][i]['path'] = self._download(
                    spath, metasearch=False)
            file['path'] = npath
        else:
            file = {}
            file['metadata'] = self._meta(url, prompt=prompt)
            if secondaryFiles:
                file['secondaryFiles'] = self._get_secondary_files(secondaryFiles,
                                                                   url,
                                                                   prompt=prompt)
            file['path'] = npath
            if prompt:
                self._rbx_dump(file)
        return file

    def _download(self, url, metasearch=True):
        if url.startswith('data:,'):
            return self._data_url(url)
        if '://' not in url:
            url = 'file://' + url
        if url.startswith('file://'):
            return self._local(url)

        log.debug('Downloading %s', url)
        r = requests.get(url)
        try:
            r.raise_for_status()
        except Exception as e:
            log.error('Resource unavailable %s: ', url)
            raise ResourceUnavailable(str(e))
        dest = self._get_dest_for_url(url)
        with open(dest, 'wb') as fp:
            for chunk in r.iter_content(chunk_size=1024):
                fp.write(chunk)
        if metasearch:
            meta = self._get_meta_for_url(url)
            if meta:
                with open(dest + '.rbx.json', 'w') as fp:
                    to_json(meta, fp)
        return os.path.abspath(dest)

    def _get_meta_for_url(self, url):
        log.debug('Fetching metadata for %s', url)
        chunks = list(urlparse(url))
        chunks[2] += '.rbx.json'
        meta_url = urlunparse(chunks)
        r = requests.get(meta_url)
        if not r.ok:
            log.warning('Failed to get metadata for URL %s', url)
            return
        try:
            meta = r.json()
            assert isinstance(meta, dict)
            log.info('Fetched metadata from %s', meta_url)
        except:
            log.warning('Metadata not valid JSON object: %s', meta_url)
            return
        return meta

    def _meta(self, path, prompt=True):
        file_meta = {}
        if os.path.exists(path + '.meta'):
            with open(path + '.meta') as m:
                file_meta = json.load(m)
        elif prompt:
            file_meta = self._metadata_prompt(os.path.basename(path))
        return file_meta

    def _metadata_prompt(self, input, metadata=None):
        cont = ''
        if metadata is None:
            metadata = {}
            cont = 'n'
            # cont = raw_input('Metadata for file %s not found. '
            #                  'Do you want to set it manually? [Y/n] '
            #                  % input)
            # print(cont.__class__)
            # cont = cont.lower().strip()
        if cont == 'y' or cont == '':
            key = raw_input("Key: ")
            if key == '':
                return metadata
            value = raw_input("Value: ")
            if value == '':
                return metadata
            metadata[key] = value
            return self._metadata_prompt(input, metadata)
        return metadata

    def _get_secondary_files(self, secondaryFiles, input, autodetect=True,
                             prompt=True):

        def secondary_files_autodetect(path):
            log.info('Searching for additional files for file: %s', path)
            return [fn for fn in glob.glob(path + '.*')
                    if not (os.path.basename(fn).endswith('.meta') or
                            os.path.basename(fn).endswith('.rbx.json'))
                    ]

        secFiles = []
        if secondaryFiles:
            for n, sf in enumerate(secondaryFiles):
                path = sec_files_naming_conv(input, sf)
                log.info('Downloading: %s', path)
                secFiles.append({'path': self._download(
                    path, metasearch=False)})
        if autodetect:
            if not secondaryFiles:
                secondaryFiles = []
            autodetected = secondary_files_autodetect(input)
            ad = []
            names = [os.path.basename(f) for f in secondaryFiles]
            for sf in autodetected:
                sf = sf.replace(input, '')
                if sf not in names:
                    ad.append(str(sf))
            if ad:
                cont = raw_input('Do you want to include autodetected '
                                 'additional files for file %s %s? [Y/n] '
                                 % (input, str(ad))).lower().strip()
                if cont == 'y' or cont == '':
                    log.info("Additional files %s included" % str(ad))
                    secFiles.extend(self._get_secondary_files(
                        ad, input, autodetect=False, prompt=False))
        if prompt:
            prompt = self._prompt_files(input, secFiles=secondaryFiles)
            secFiles.extend(self._get_secondary_files(
                prompt, input, autodetect=False, prompt=False))
        return secFiles

    def _data_url(self, url):
        data = url[len('data:,'):]
        dest = tempfile.mktemp(dir=self.task_dir)
        with open(dest, 'w') as fp:
            fp.write(data)
        return os.path.abspath(dest)

    def _get_dest_for_url(self, url):
        path = urlparse(url).path
        name = path.split('/')[-1]
        tgt = os.path.join(self.task_dir, name)
        if os.path.exists(tgt) or not name:
            return tempfile.mktemp(dir=self.task_dir)
        return tgt

    def _local(self, url):
        path = url[len('file://'):]
        if not os.path.exists(path):
            raise ResourceUnavailable('Not a file: %s' % path)
        return os.path.abspath(path)

    def _prompt_files(self, input, prompt=None, secFiles=None):
        cont = ''
        if prompt is None:
            prompt = []
            cont = raw_input('Do you want to include some additional '
                             'files for file %s %s? [y/N] '
                             % (input, str(secFiles))).lower().strip()
        if cont == 'y':
            ext = raw_input('Extension: ')
            if ext == '':
                return prompt
            elif ext in secFiles:
                log.error('Extension %s already included' % ext)
            else:
                prompt.append(ext)
                secFiles.append(ext)
            return self._prompt_files(input, prompt, secFiles=secFiles)
        else:
            return prompt

    def _rbx_dump(self, input):
        cont = raw_input('Do you want to create rbx.json for file %s? [Y/n] ' % input['path']).lower().strip()
        if cont == 'y' or cont == '':
            filename = input['path'] + '.rbx.json'
            with open(filename, 'w') as f:
                json.dump(input, f)
                log.info('File %s created.', filename)



#!/usr/bin/env python

import json
import argparse
import os

from rabix.common.ref_resolver import from_url
from rabix.cli.adapter import CLIJob


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("tool", type=str)
    parser.add_argument("job_order", type=str)
    parser.add_argument("--conformance-test", action="store_true")
    parser.add_argument("--basedir", type=str)
    parser.add_argument("--no-container", action="store_true")
    args = parser.parse_args()

    tool = from_url(args.tool)
    job = from_url(args.job_order)

    def path_mapper(path):
        return os.path.normpath(os.path.join(args.basedir or '.', path))

    cli_job = CLIJob(job, tool, path_mapper)
    print(json.dumps({
        'args': cli_job.make_arg_list(),
        'stdin': cli_job.stdin,
        'stdout': cli_job.stdout,
    }))


if __name__ == "__main__":
    main()



import os


def validate_inputs(tool, job):
    required = tool.get('inputs', {}).get('required', [])
    inputs = job.get('inputs', None)
    for req in required:
        if req not in inputs.keys():
            raise Exception('Required input not set')
        else:
            if isinstance(inputs[req], list):
                for f in inputs[req]:
                    if not os.path.exists(f['path']):
                        raise Exception("File %s doesn't exist" % f['path'])
            else:
                if not os.path.exists(inputs[req]['path']):
                    raise Exception(
                        "File %s doesn't exist" % inputs[req]['path'])



from __future__ import print_function
import os
import docopt
import sys
import logging
import six

from six.moves.urllib.parse import urlparse

from rabix import __version__ as version
from rabix.common.util import log_level, dot_update_dict
from rabix.common.models import Job, IO, File
from rabix.common.context import Context
from rabix.common.ref_resolver import from_url
from rabix.common.errors import RabixError
from rabix.cli.adapter import CLIJob
from rabix.executor import Executor
from rabix.cli import CliApp

import rabix.cli
import rabix.docker
import rabix.expressions
import rabix.workflows
import rabix.schema


TEMPLATE_RESOURCES = {
    "cpu": 4,
    "mem": 5000
}


TEMPLATE_JOB = {
    'app': 'http://example.com/app.json',
    'inputs': {},
    'platform': 'http://example.org/my_platform/v1',
    'allocatedResources': TEMPLATE_RESOURCES
}

USAGE = '''
Usage:
    rabix <tool> [-v...] [-hcI] [-d <dir>] [-i <inp>] [{resources}] [-- {inputs}...]
    rabix --version

    Options:
  -d --dir=<dir>       Working directory for the task. If not provided one will
                       be auto generated in the current dir.
  -h --help            Show this help message. In conjunction with tool,
                       it will print inputs you can provide for the job.

  -I --install         Only install referenced tools. Do not run anything.
  -i --inp-file=<inp>  Inputs
  -c --print-cli       Only print calculated command line. Do not run anything.
  -v --verbose         Verbosity. More Vs more output.
     --version         Print version and exit.
'''

TOOL_TEMPLATE = '''
Usage:
  tool {inputs}
'''


def make_resources_usage_string(template=TEMPLATE_RESOURCES):
    param_str = []
    for k, v in six.iteritems(template):
        if type(v) is bool:
            arg = '--resources.%s' % k
        else:
            arg = '--resources.%s=<%s>' % (k, type(v).__name__)
        param_str.append(arg)
    return ' '.join(param_str)


TYPE_MAP = {
    'Job': Job.from_dict,
    'IO': IO.from_dict,
    'File': File.from_dict
}


def init_context():
    executor = Executor()
    context = Context(TYPE_MAP, executor)

    for module in (
            rabix.cli, rabix.expressions, rabix.workflows,
            rabix.schema, rabix.docker
    ):
        module.init(context)

    return context


def fix_types(tool):
    requirements = tool.get('requirements', {})
    environment = requirements.get('environment')

    # container type
    if (environment and
            isinstance(environment.get('container'), dict) and
            environment['container'].get('type') == 'docker'):
        environment['container']['@type'] = 'Docker'

    # tool type
    if '@type' not in tool:
        tool['@type'] = 'CommandLine'

    if tool['@type'] == 'Workflow':
        for step in tool['steps']:
            fix_types(step['app'])

    # schema type
    inputs = tool.get('inputs')
    if isinstance(inputs, dict) and '@type' not in inputs:
        inputs['@type'] = 'JsonSchema'

    outputs = tool.get('outputs')
    if isinstance(outputs, dict) and '@type' not in outputs:
        outputs['@type'] = 'JsonSchema'


def make_app_usage_string(app, template=TOOL_TEMPLATE, inp=None):

    inp = inp or {}

    def resolve(k, v, usage_str, param_str, inp):

        if v.constructor == TYPE_MAP['File']:
            arg = '--%s=<file>' % k
            to_append = usage_str
        else:
            arg = '--%s=<%s>' % (k, v.constructor.__name__)
            to_append = param_str

        if v.depth > 0:
            arg += '... '

        if not v.required or v.id in inp:
            arg = '['+arg+']'

        to_append.append(arg)

    def resolve_object(obj, usage_str, param_str, inp, root=False):
        properties = obj.inputs.io if root else obj.objects
        for input in properties:
            key = input.id if root else '.'.join([obj.id, input.id])
            resolve(key, input, usage_str, param_str, inp.keys())

    usage_str = []
    param_str = []

    resolve_object(app, usage_str, param_str, inp, root=True)
    usage_str.extend(param_str)
    return template.format(resources=make_resources_usage_string(),
                           inputs=' '.join(usage_str))


def resolve_path(basedir, url):
    if basedir and url.scheme == 'file' and not os.path.isabs(url.path):
        l = list(url)  # transform from tuple
        l[2] = os.path.join(basedir, url.path)  # update path
        url = url.__class__(*l)  # return back to original type

    return url


def construct_value(constructor, val, basedir):
    if constructor == TYPE_MAP['File']:
        url = urlparse(val, 'file')
        url = resolve_path(basedir, url)
        val = {'path': url}

    return constructor(val)


def resolve_values(inp, nval, inputs, basedir=None):
    if isinstance(nval, list):
        inputs[inp.id] = [construct_value(inp.constructor, nv, basedir)
                          for nv in nval]
    else:
        inputs[inp.id] = construct_value(inp.constructor, nval, basedir)


def get_inputs_from_file(app, args, startdir):
    inp = {}
    inputs = app.inputs.io
    resolve_nested_paths(inp, inputs, args, startdir)
    return {'inputs': inp}


def resolve_nested_paths(inp, inputs, args, startdir):
    for input in inputs:
        nval = args.get(input.id)
        if nval:
            if input.depth > 0 and input.constructor == dict:
                inp[input.id] = []
                for sk, sv in enumerate(nval):
                    inp[input.id].append({})
                    resolve_nested_paths(
                        inp[input.id][sk],
                        input.objects,
                        args[input.id],
                        startdir
                    )
            else:
                resolve_values(input, nval, inp, startdir)


def get_inputs(app, args):
    inputs = {}
    properties = app.inputs.io
    for input in properties:
        nval = args.get('--' + input.id) or args.get(input.id)
        if nval:
            resolve_values(input, nval, inputs)
    return {'inputs': inputs}


def get_tool(args):
    if args['<tool>']:
        return from_url(args['<tool>'])


def dry_run_parse(args=None):
    args = args or sys.argv[1:]
    args += ['an_input']
    usage = USAGE.format(resources=make_resources_usage_string(),
                         inputs='<inputs>')
    try:
        return docopt.docopt(usage, args, version=version, help=False)
    except docopt.DocoptExit:
        return


def main():
    logging.basicConfig(level=logging.WARN)
    if len(sys.argv) == 1:
        print(USAGE)
        return

    usage = USAGE.format(resources=make_resources_usage_string(),
                         inputs='<inputs>')
    app_usage = usage

    if len(sys.argv) == 2 and \
            (sys.argv[1] == '--help' or sys.argv[1] == '-h'):
        print(USAGE)
        return

    dry_run_args = dry_run_parse()
    if not dry_run_args:
        print(USAGE)
        return

    if not (dry_run_args['<tool>']):
        print('You have to specify a tool, with --tool option')
        print(usage)
        return

    tool = get_tool(dry_run_args)
    if not tool:
        print("Couldn't find tool.")
        return

    fix_types(tool)

    context = init_context()
    app = context.from_dict(tool)

    if dry_run_args['--install']:
        app.install()
        print("Install successful.")
        return

    try:
        args = docopt.docopt(usage, version=version, help=False)
        job = TEMPLATE_JOB
        logging.root.setLevel(log_level(dry_run_args['--verbose']))

        if args['--inp-file']:
            startdir = os.path.dirname(args.get('--inp-file'))
            input_file = from_url(args.get('--inp-file'))
            dot_update_dict(
                job['inputs'],
                get_inputs_from_file(app, input_file, startdir)['inputs']
            )

        app_inputs_usage = make_app_usage_string(
            app, template=TOOL_TEMPLATE, inp=job['inputs'])

        app_usage = make_app_usage_string(app, USAGE, job['inputs'])

        app_inputs = docopt.docopt(app_inputs_usage, args['<inputs>'])

        if args['--help']:
            print(app_usage)
            return

        inp = get_inputs(app, app_inputs)
        job['inputs'].update(inp['inputs'])

        if args['--print-cli']:
            if not isinstance(app, CliApp):
                print(dry_run_args['<tool>'] + " is not a command line app")
                return
            job['@id'] = args.get('--dir')
            job['app'] = app.to_dict(context)
            j = Job.from_dict(context, job)
            adapter = CLIJob(j.to_dict(context), j.app)
            print(adapter.cmd_line())
            return

        job['@id'] = args.get('--dir')
        job['app'] = app.to_dict(context)

        try:
            context.executor.execute(Job.from_dict(context, job),
                                     lambda _, result: print(result))
        except RabixError as err:
            print(err.message)

    except docopt.DocoptExit:
        print(app_usage)
        return


if __name__ == '__main__':
    main()



import os
import six
import shlex
import logging
import docker

from docker.errors import APIError
from six.moves.urllib.parse import urlparse

from rabix.cli.cli_app import Container
from rabix.docker.container import get_image
from rabix.common.errors import ResourceUnavailable


log = logging.getLogger(__name__)

DOCKER_DEFAULT_API_VERSION = "1.12"
DOCKER_DEFAULT_TIMEOUT = 60


def docker_client(docker_host=None,
                  api_version=DOCKER_DEFAULT_API_VERSION,
                  timeout=DOCKER_DEFAULT_TIMEOUT,
                  tls=None):

    docker_host = docker_host or os.getenv("DOCKER_HOST", None)
    tls = False if tls is None else os.getenv("DOCKER_TLS_VERIFY", "") != ""
    docker_cert_path = os.getenv("DOCKER_CERT_PATH", "")  # ???

    return docker.Client(docker_host, api_version, timeout, tls)


def make_config(**kwargs):
    keys = ['Hostname', 'Domainname', 'User', 'Memory', 'MemorySwap',
            'CpuShares', 'Cpuset', 'AttachStdin', 'AttachStdout',
            'AttachStderr', 'PortSpecs', 'ExposedPorts', 'Tty', 'OpenStdin',
            'StdinOnce', 'Env', 'Cmd', 'Image', 'Volumes', 'WorkingDir',
            'Entrypoint', 'NetworkDisabled', 'OnBuild']
    cfg = {
        'AttachStdin': False,
        'AttachStdout': False,
        'AttachStderr': False,
        'Tty': False,
        'Privileged': False,
        }
    cfg.update({k[0].upper() + k[1:]: v for k, v in six.iteritems(kwargs)})
    cfg = {k: v for k, v in six.iteritems(cfg) if k in keys}
    entrypoint = cfg.get("Entrypoint")
    if isinstance(entrypoint, six.string_types):
        cfg['Entrypoint'] = shlex.split(entrypoint)
    return cfg


class BindDict(dict):
    def items(self):
        ret = []
        for k, v in six.iteritems(super(BindDict, self)):
            ret.append((v, k))
        return ret


class DockerContainer(Container):

    def __init__(self, uri, image_id, dockr=None):
        super(DockerContainer, self).__init__()
        self.uri = uri.lstrip("docker://")
        self.image_id = image_id
        self.docker_client = dockr or docker_client()
        self.config = {}
        self.volumes = {}
        self.binds = {}

    def install(self, *args, **kwargs):
        try:
            get_image(self.docker_client,
                      image_id=self.image_id,
                      repo=self.uri)
        except APIError as e:
            if e.response.status_code == 404:
                log.info('Image %s not found:' % self.image_id)
                raise RuntimeError('Image %s not found:' % self.image_id)

    def prepare_paths(self, job):
        self.volumes, self.binds = self._volumes(job)

    def _volumes(self, job):
        remaped_job = job.inputs
        volumes = {}
        binds = {}

        inputs = job.app.inputs.schema['properties']
        input_values = remaped_job
        self._remap(inputs, input_values, volumes, binds, remaped_job)

        return volumes, BindDict(binds)

    def _remap(self, inputs, input_values, volumes, binds, remaped_job,
               parent=None):
        is_single = lambda i: any([inputs[i]['type'] == 'directory',
                                   inputs[i]['type'] == 'file'])
        is_array = lambda i: inputs[i]['type'] == 'array' and any([
            inputs[i]['items']['type'] == 'directory',
            inputs[i]['items']['type'] == 'file'])
        is_object = lambda i: (inputs[i]['type'] == 'array' and
                               inputs[i]['items']['type'] == 'object')
        if inputs:
            single = filter(is_single, [i for i in inputs])
            lists = filter(is_array, [i for i in inputs])
            objects = filter(is_object, [i for i in inputs])
            for inp in single:
                self._remap_single(inp, input_values, volumes, binds,
                                   remaped_job, parent)
            for inp in lists:
                self._remap_list(inp, input_values, volumes, binds,
                                 remaped_job, parent)
            for obj in objects:
                if input_values.get(obj):
                    for num, o in enumerate(input_values[obj]):
                        self._remap(inputs[obj]['items']['properties'],
                                    o, volumes, binds, remaped_job[obj][num],
                                    parent='/'.join([obj, str(num)]))

    def _remap_single(self, inp, input_values, volumes, binds, remaped_job,
                      parent):
        if input_values.get(inp):
            if parent:
                docker_dir = '/' + '/'.join([parent, inp])
            else:
                docker_dir = '/' + inp
            dir_name, file_name = os.path.split(
                os.path.abspath(input_values[inp]['path']))
            volumes[docker_dir] = {}
            binds[docker_dir] = dir_name
            remaped_job[inp]['path'] = '/'.join(
                [docker_dir, file_name])

    def _remap_list(self, inp, input_values, volumes, binds, remaped_job,
                    parent):
        if input_values[inp]:
            for num, inv in enumerate(input_values[inp]):
                if parent:
                    docker_dir = '/' + '/'.join([parent, inp, str(num)])
                else:
                    docker_dir = '/' + '/'.join([inp, str(num)])
                dir_name, file_name = os.path.split(
                    os.path.abspath(inv['path']))
                volumes[docker_dir] = {}
                binds[docker_dir] = dir_name
                remaped_job[inp][num]['path'] = '/'.join(
                    [docker_dir, file_name])

    def _envvars(self, job):
        envvars = (job.app.annotations or {}).get('environment', {})
        envlst = []
        for env, val in six.iteritems(envvars):
            envlst.append('='.join([env, val]))
        return envlst

    def set_config(self, *args, **kwargs):
        self.prepare_paths(kwargs.get('job'))
        user = kwargs.get('user', None) or ':'.join([str(os.getuid()),
                                                     str(os.getgid())])
        self.job_dir = kwargs.get('job_dir')
        self.volumes['/' + self.job_dir] = {}
        self.binds['/' + self.job_dir] = os.path.abspath(self.job_dir)
        self.stderr = kwargs.get('stderr') or 'out.err'
        cfg = {
            "Image": self.image_id,
            "User": user,
            "Volumes": self.volumes,
            "WorkingDir": self.job_dir
        }
        self.config = make_config(**cfg)

    def _start(self, cmd):
        self.config["Cmd"] = ['bash', '-c', cmd]
        try:
            self.container = self.docker_client.create_container_from_config(
                self.config)
        except APIError:
            raise RuntimeError('Failed to create Container')
        try:
            self.docker_client.start(container=self.container, binds=self.binds)
        except APIError:
            logging.error('Failed to run container %s' % self.container)
            raise RuntimeError('Unable to run container from image %s:'
                               % self.image_id)

    def run(self, cmd):  # should be run(self, cmd_line, job)
        self._start(cmd)
        self.get_stderr(file='/'.join([os.path.abspath(self.job_dir),
                                       self.stderr]))
        if not self.is_success():
            raise RuntimeError("err %s" % self.get_stderr())

    def remove(self, success_only=False):
        self.wait()
        if not success_only or self.is_success():
            self.docker_client.remove_container(self.container)
        return self

    def inspect(self):
        return self.docker_client.inspect_container(self.container)

    def is_running(self):
        return self.inspect()['State']['Running']

    def wait(self):
        if self.is_running():
            self.docker_client.wait(self.container)
        return self

    def is_success(self):
        return self.wait().inspect()['State']['ExitCode'] == 0

    def get_stdout(self, file=None):
        if file:
            f = open(file, 'w', buffering=1)
        if self.is_running():
            for out in self.docker_client.attach(self.container, stdout=True,
                                                 stderr=False, stream=True,
                                                 logs=True):
                if file:
                    f.write(out.rstrip() + '\n')
                else:
                    print(out.rstrip())
            if file:
                f.close()
        else:
            print(self.docker_client.logs(self.container))
        return self

    def get_stderr(self, file=None):
        if file:
            f = open(file, 'w')
            if self.is_running():
                for out in self.docker_client.attach(self.container,
                                                     stdout=False, stderr=True,
                                                     stream=True, logs=True):
                    if file:
                        f.write(str(out).rstrip() + '\n')
                    else:
                        print(out.rstrip())
            else:
                f.write(self.docker_client.logs(self.container,
                                                stdout=False, stderr=True))
            f.close()
        else:
            print(self.docker_client.logs(self.container))
        return self

    def to_dict(self, context=None):
        return {
            "@type": "Docker",
            "type": "docker",
            "uri": self.uri,
            "imageId": self.image_id
        }

    @classmethod
    def from_dict(cls, context, d):
        return cls(d['uri'], d.get('imageId'))



import os
import six
import json
import stat

from rabix.cli.adapter import CLIJob
from rabix.common.models import App, File
from rabix.common.io import InputCollector


class Resources(object):

    def __init__(self, cpu, mem):
        self.cpu = cpu
        self.mem = mem

    def to_dict(self, serializer=None):
        return {
            "@type": "Resources",
            "cpu": self.cpu,
            "mem": self.mem
        }

    @classmethod
    def from_dict(cls, d):
        cls(d.get('cpu', 0), d.get('mem', 0))


class Container(object):

    def __init__(self):
        self.inputCollector = InputCollector()

    def install(self, job):
        pass

    def set_config(self, *args, **kwargs):
        pass

    def ensure_files(self, job, job_dir):
        '''
        Resolve paths of all input files
        '''
        inputs = job.app.inputs.io
        self.inputCollector.set_dir(job_dir)
        input_values = job.inputs
        self._resolve(inputs, input_values, job.inputs)

    def _resolve(self, inputs, input_values, job):
        is_single = lambda i: i.constructor == File.from_dict and i.depth == 0
        is_array = lambda i: i.constructor == File.from_dict and i.depth > 0
        # is_object = lambda i: i.constructor == 'array' and i.itemType == 'object'

        if inputs:
            single = filter(is_single, [i for i in inputs])
            lists = filter(is_array, [i for i in inputs])
            # objects = filter(is_object, [i for i in inputs])
            for inp in single:
                ival = input_values.get(inp.id)
                if ival:
                    job[inp.id] = self._resolve_file(
                        inp.annotations.get('secondaryFiles'),
                        ival
                    )

            for inp in lists:
                ivals = input_values.get(inp.id)
                if ivals:
                    job[inp.id] = [self._resolve_file(
                        inp.annotations.get('secondaryFiles'),
                        ival
                    ) for ival in ivals]
            # for obj in objects:
            #     if input_values.get(obj.id):
            #         for num, o in enumerate(input_values[obj.id]):
            #             self._resolve(obj.objects, o, job[obj.id][num])

    def _resolve_file(self, secondary_files, input_value):

        # if input_value['path'].endswith('.rbx.json'):
        #     job[inp.id] = from_url(input_value['path'])
        # else:
        url = input_value.url
        if url.scheme == 'file':
            url = url.path
        else:
            url = url.geturl()
        return self.inputCollector.download(url, secondary_files)


class Requirements(object):

    def __init__(self, container=None, resources=None, platform_features=None):
        self.container = container
        self.resources = resources
        self.platform_features = platform_features

    def to_dict(self, context):
        return {
            "@type": "Requirements",
            "environment": {"container": self.container.to_dict(context)},
            "resources": context.to_dict(self.resources),
            "platformFeatures": self.platform_features
        }

    @classmethod
    def from_dict(cls, context, d):
        return cls(
            context.from_dict(d.get('environment', {}).get('container')),
            context.from_dict(d.get('resources')),
            context.from_dict(d.get('platformFeatures'))
        )


class CliApp(App):
    WORKING_DIR = '/work'

    def __init__(self, app_id, inputs, outputs, context,
                 app_description=None,
                 annotations=None,
                 platform_features=None,
                 adapter=None,
                 software_description=None,
                 requirements=None):
        super(CliApp, self).__init__(
            app_id, inputs, outputs,
            app_description=app_description,
            annotations=annotations,
            platform_features=platform_features
        )
        self.adapter = adapter
        self.software_description = software_description
        self.requirements = requirements
        self.context = context

    def run(self, job, job_dir=None):
        job_dir = job_dir or job.id
        os.mkdir(job_dir)
        os.chmod(job_dir, os.stat(job_dir).st_mode | stat.S_IROTH |
                 stat.S_IWOTH)
        if self.requirements.container:
            self.ensure_files(job, job_dir)
            self.install(job=job)
            self.job_dump(job, job_dir, self.context)
            self.set_config(job=job, job_dir=job_dir)
            adapter = CLIJob(job.to_dict(self.context), job.app)
            cmd_line = adapter.cmd_line()
            self.requirements.container.run(cmd_line)
            with open(os.path.abspath(job_dir) + '/result.cwl.json', 'w') as f:
                outputs = adapter.get_outputs(os.path.abspath(job_dir))
                for k, v in six.iteritems(outputs):
                    if v:
                        with open(v['path'] + '.rbx.json', 'w') as rx:
                            json.dump(v, rx)
                json.dump(outputs, f)
                return outputs

    def install(self, *args, **kwargs):
        if self.requirements and self.requirements.container:
            self.requirements.container.install(*args, **kwargs)

    def ensure_files(self, job, job_dir):
        if self.requirements and self.requirements.container:
            self.requirements.container.ensure_files(job, job_dir)

    def set_config(self, *args, **kwargs):
        if self.requirements and self.requirements.container:
            self.requirements.container.set_config(*args, **kwargs)

    def to_dict(self, context=None):
        d = super(CliApp, self).to_dict(context)
        d.update({
            "@type": "CommandLine",
            'name': self.id,
            'adapter': self.adapter,
            'annotations': self.annotations,
            'platform_features': self.platform_features,
            'inputs': self.inputs.schema,
            'outputs': self.outputs.schema,
            'requirements': self.requirements.to_dict(context)
        })
        return d

    @classmethod
    def from_dict(cls, context, d):
        return cls(d.get('@id', d.get('name')),
                   context.from_dict(d['inputs']),
                   context.from_dict(d['outputs']),
                   context,
                   app_description=d.get('appDescription'),
                   annotations=d.get('annotations'),
                   platform_features=d.get('platform_features'),
                   adapter=context.from_dict(d.get('adapter')),
                   software_description=d.get('softwareDescription'),
                   requirements=Requirements.from_dict(
                       context, d.get('requirements', {})))



import os
import sys
import json
import shutil

from nose.tools import raises

from rabix.tests import mock_app_bad_repo, mock_app_good_repo
from rabix.main import main
from rabix.docker import docker_client, get_image


@raises(Exception)
def test_provide_image_bad_repo():
    uri = mock_app_bad_repo["tool"]["requirements"]["environment"][
        "container"]["uri"]
    imageId = mock_app_bad_repo["tool"]["requirements"]["environment"][
        "container"]["imageId"]
    docker = docker_client()
    get_image(docker, image_id=imageId, repo=uri)


def test_provide_image_good_repo():
    uri = mock_app_good_repo["tool"]["requirements"]["environment"][
        "container"]["uri"]
    imageId = mock_app_good_repo["tool"]["requirements"]["environment"][
        "container"]["imageId"]
    docker = docker_client()
    get_image(docker, image_id=imageId, repo=uri)


def test_expr_and_meta():
    sys.argv = ['rabix', './rabix/tests/test-expr/bwa-mem.json',
                '-i', './rabix/tests/test-cmdline/inputs.json',
                '--dir', 'test1', '--']
    main()
    with open(os.path.abspath('./test1') + '/output.sam.rbx.json') as m:
        meta = json.load(m)
        assert meta['metadata']['expr_test'] == 'successful'
    shutil.rmtree(os.path.abspath('./test1'))
    sys.argv = ['rabix', '-i', './rabix/tests/test-cmdline/inputs.json',
                './rabix/tests/test-expr/bwa-mem.json',
                '--dir', 'test2']
    main()
    with open(os.path.abspath('./test2') + '/output.sam.rbx.json') as m:
        meta = json.load(m)
        assert meta['metadata']['expr_test'] == 'successful'
    shutil.rmtree(os.path.abspath('./test2'))


def test_fetch_remote_files():
    sys.argv = ['rabix', '--dir', 'test_fetch_remote',
                './rabix/tests/test-cmdline/bwa-mem.json#tool', '--',
                '--reads',
                'https://s3.amazonaws.com/rabix/rabix-test/'
                'example_human_Illumina.pe_1.fastq', '--reads',
                'https://s3.amazonaws.com/rabix/rabix-test/'
                'example_human_Illumina.pe_2.fastq', '--reference',
                './rabix/tests/test-files/chr20.fa']
    main()
    assert os.path.exists(os.path.abspath('./test_fetch_remote') + '/output.sam')
    shutil.rmtree(os.path.abspath('./test_fetch_remote'))


def test_params_from_input_file():
    sys.argv = ['rabix', '-i', 'rabix/tests/test-cmdline/inputs.json',
                'rabix/tests/test-expr/bwa-mem.json',
                '-d', 'test_from_input_file']
    main()
    assert os.path.exists(os.path.abspath('./test_from_input_file') + '/output.sam')
    shutil.rmtree(os.path.abspath('./test_from_input_file'))


def test_override_input():
    sys.argv = ['rabix', '-i', 'rabix/tests/test-cmdline/inputs.json', '-d',
                'test_override_input', 'rabix/tests/test-expr/bwa-mem.json', '--',
                '--reference', 'rabix/tests/test-files/chr20.fa']
    main()
    assert os.path.exists(os.path.abspath('./test_override_input') + '/output.sam')
    shutil.rmtree(os.path.abspath('./test_override_input'))



import os
from xdg.BaseDirectory import save_config_path, xdg_data_dirs
from yapsy.ConfigurablePluginManager import ConfigurablePluginManager
from yapsy.VersionedPluginManager import VersionedPluginManager
from yapsy.PluginManager import PluginManagerSingleton
from yapsy.IPlugin import IPlugin
from six.moves.configparser import SafeConfigParser


class ExpressionEvalPlugin(IPlugin):

    def __init__(self):
        super(ExpressionEvalPlugin, self).__init__()

    def evaluate(self, expression=None, job=None, context=None, *args,
                 **kwargs):
        raise RuntimeError('Not implemented')


class Evaluator(object):

    APP_NAME = 'expression-evaluators'
    _default_dir = 'evaluators'

    def __init__(self, plugin_dir=None):
        self.config = SafeConfigParser()
        config_path = save_config_path(self.APP_NAME)
        self.config_file = os.path.join(config_path, self.APP_NAME + ".conf")
        self.config.read(self.config_file)

        this_dir = os.path.abspath(os.path.dirname(__file__))
        self.plugin_dir = plugin_dir or os.path.join(
            this_dir, self._default_dir)
        places = [self.plugin_dir, ]
        [places.append(os.path.join(path, self.APP_NAME, "evaluators")) for
         path in xdg_data_dirs]

        PluginManagerSingleton.setBehaviour([
            ConfigurablePluginManager,
            VersionedPluginManager,
        ])

        self.manager = PluginManagerSingleton.get()
        self.manager.setConfigParser(self.config, self.write_config)
        self.manager.setPluginInfoExtension("expr-plugin")
        self.manager.setPluginPlaces(places)
        self.manager.collectPlugins()

    def _get_all_evaluators(self):
        return self.manager.getAllPlugins()

    def _get_evaluator(self, name):
        pl = self.manager.getPluginByName(name)
        if not pl:
            raise Exception('No expression evaluator %s' % name)
        return pl.plugin_object

    def write_config(self):
        f = open(self.config_file, "w")
        self.config.write(f)
        f.close()

    def evaluate(self, lang, expression, *args, **kwargs):
        pl = self._get_evaluator(lang)
        return pl.evaluate(expression, *args, **kwargs)



import logging
import six

from nose.tools import nottest, assert_equal

from os.path import abspath, join
from rabix.common.ref_resolver import from_url
from rabix.main import init_context, fix_types

logging.basicConfig(level=logging.DEBUG)


@nottest
def assert_execution(job, outputs):
    result = job.run()
    assert_equal(result, outputs)


def test_workflow():
    path = abspath(join(__file__, '../../test_runtime/wf_tests.yaml'))

    context = init_context()
    doc = from_url(path)
    tests = doc['tests']
    for test_name, test in six.iteritems(tests):
        features = test.get('requiresFeatures', [])
        if 'map' not in features:
            fix_types(test['job']['app'])
            yield assert_execution, context.from_dict(test['job']), test['outputs']



from rabix.docker.docker_app import DockerContainer, docker_client
from rabix.docker.container import find_image, get_image


def init(context):
    context.add_type('Docker', DockerContainer.from_dict)



from .cli_app import CliApp, Requirements


def init(context):
    context.add_type('CommandLine', CliApp.from_dict)
    context.add_type('CliApp', CliApp.from_dict)
    context.add_type('CliTool', CliApp.from_dict)
    context.add_type('Requirements', Requirements.from_dict)



import io
import sys

from os.path import dirname
from setuptools import setup, find_packages

sys.path.append(dirname(__file__))
from rabix import __version__

requires = [
    x.strip() for x in
    io.open('requirements.txt')
]

setup(
    name="rabix",
    version=__version__,
    packages=find_packages(),
    entry_points={
        'console_scripts': ['rabix = rabix.main:main',
                            'rabix-tools = rabix.tools.cli:main'],
    },
    install_requires=requires,
    package_data={'': ['*.expr-plugin']},
    long_description=io.open('README.md').read(),
    zip_safe=False,
    test_suite='tests',
    license='AGPLv3',

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'License :: OSI Approved :: ' +
        'GNU Affero General Public License v3 or later (AGPLv3+)'
    ]
)



import copy
import six
import os

from rabix.common.util import wrap_in_list
from rabix.common.errors import RabixError
from rabix.common.models import Job


class Executor(object):

    def __init__(self):
        pass

    @staticmethod
    def depth(val):
        d = 0
        cur = val
        while isinstance(cur, list):
            if not cur:
                break
            cur = cur[0]
            d += 1

        return d

    @staticmethod
    def split_job(job):
        parallel_input = None
        for input_name, input_val in six.iteritems(job.inputs):
            io = job.app.get_input(input_name)
            val_d = Executor.depth(input_val)
            if val_d == io.depth:
                continue
            if val_d > io.depth + 1:
                raise RabixError("Depth difference to large")
            if val_d < io.depth:
                raise RabixError("Insufficient dimensionality")
            if parallel_input:
                raise RabixError("Already parallelized by input '%%s'" % parallel_input)

            parallel_input = input_name

        if parallel_input:
            jobs = []
            for i, val in enumerate(job.inputs[parallel_input]):
                inputs = copy.deepcopy(job.inputs)
                inputs[parallel_input] = val
                jobs.append(Job(job.id+"_"+str(i), job.app, inputs, {}))
            return jobs
        else:
            return job

    def execute(self, job, callback=None, callback_id=None):
        # TODO: resources, instances, scheduling, yada yada...
        jobs = self.split_job(job)
        result = None
        if isinstance(jobs, list):
            results = []
            for job in jobs:
                results.append(job.run())

            combined = {}
            for result in results:
                for k, v in six.iteritems(result):
                    acc = combined.get(k, [])
                    acc.append(v)
                    combined[k] = acc
            result = combined
        else:
            result = jobs.run()

        if callback:
            callback(callback_id, result)




infinite_loop = {
    "tool": {
        "softwareDescription": {},
        "documentAuthor": "boysha",
        "requirements": {
            "environment": {
                "container": {
                    "type": "docker",
                    "uri": "docker:infinite_loop#latest",
                    "imageId": "e678dddee492"
                }
            },
            "resources": {},
            "platformFeatures": []
        },
        "inputs": {},
        "outputs": {},
        "adapter": {
            "baseCmd": [],
            "stdout": "output.sam",
            "args": []
        }
    }
}

infinite_read = {
    "tool": {
        "softwareDescription": {},
        "documentAuthor": "boysha",
        "requirements": {
            "environment": {
                "container": {
                    "type": "docker",
                    "uri": "docker:infinite_read#latest",
                    "imageId": "39be8b7d2a61"
                }
            },
            "resources": {},
            "platformFeatures": []
        },
        "inputs": {
            "input": {
                "required": "Yes",
                "minItems": 1,
                "maxItems": 2,
                "type": "file",
                "adapter": {
                    "streamable": "Yes"
                }
            },
        },
        "outputs": {},
        "adapter": {
            "baseCmd": [],
            "stdout": "output.sam",
            "args": []
        }
    }
}

mock_app_good_repo = {
    "tool": {
        "softwareDescription": {},
        "documentAuthor": "boysha",
        "requirements": {
            "environment": {
                "container": {
                    "type": "docker",
                    "uri": "ubuntu:14.10",
                    "imageId": "277eb4304907"
                }
            },
            "resources": {},
            "platformFeatures": []
        },
        "inputs": {},
        "outputs": {},
        "adapter": {
            "baseCmd": [],
            "stdout": "output.sam",
            "args": []
        }
    }
}

mock_app_bad_repo = {
    "tool": {
        "softwareDescription": {},
        "documentAuthor": "boysha",
        "requirements": {
            "environment": {
                "container": {
                    "type": "docker",
                    "uri": "docker://wrongrepository#latest",
                    "imageId": "wrongid"
                }
            },
            "resources": {},
            "platformFeatures": []
        },
        "inputs": {},
        "outputs": {},
        "adapter": {
            "baseCmd": [],
            "stdout": "output.sam",
            "args": []
        }
    }
}



import os
from os.path import splitext

from rabix.common.ref_resolver import from_url, resolve_pointer

# TODO: lists, transforms


ARGUMENT_TYPES = {}


def make_argument(arg, input_name=None):

    if input_name:
        arg_class = ARGUMENT_TYPES.get(arg.get('type'), InputArgument)
        return arg_class(input_name, arg)
    else:
        return ArgumentAdapter(arg)


class ArgumentAdapter(object):
    def __init__(self, adapter):
        self.adapter = adapter
        self.value = adapter.get('value')
        self.transforms = {
            'strip_ext': lambda x: splitext(x)[0]
        }

    def bind(self, job):
        value_from = self.adapter.get('valueFrom')
        if value_from:
            self.value = resolve_pointer(job, value_from[1:], None)
        return self

    def _cli(self, prefix, separator, value):
        transform = self.transforms.get(self.adapter.get('transform'), str)
        value = transform(value)

        if separator != ' ':
            return [prefix + separator + value]

        if prefix:
            return [prefix, value]

        return [value]

    def cli(self):
        return self._cli(self.prefix, self.separator, self.value)

    @property
    def weight(self):
        return self.adapter.get('order', 99)

    @property
    def prefix(self):
        return self.adapter.get('prefix') or ''

    @property
    def separator(self):
        return self.adapter.get('separator') or ' '


class ObjectHandler(object):

    def __init__(self, name, schema):
        self.name = name
        self.schema = schema


class InputArgument(ArgumentAdapter):

    def __init__(self, name, schema):
        adapter = schema.get('adapter', {})
        super(InputArgument, self).__init__(adapter)
        self.schema = schema
        self.name = name

    def bind(self, job):
        if not self.value:
            self.value = resolve_pointer(job, "inputs/" + self.name, None)
        return self


class FileArgument(InputArgument):
    def __init__(self, name, schema):
        super(FileArgument, self).__init__(name, schema)

    def cli(self):
        return super(FileArgument, self)._cli(
            self.prefix,
            self.separator,
            self.value['path'])
ARGUMENT_TYPES['file'] = FileArgument


class ArrayArgument(InputArgument):
    def __init__(self, name, schema):
        super(ArrayArgument, self).__init__(name, schema)

    def cli(self):
        list_separator = self.adapter.get('item_separator')
        if list_separator:
            item_arg = make_argument(self.schema['items'], 'name')
            values = []
            for val in self.value:
                item_arg.value = val
                values += item_arg.cli()

            return self._cli(self.prefix,
                             self.separator,
                             list_separator.join(values))
        else:
            items = self.schema['items']
            if not items.get('adapter'):
                items['adapter'] = {}

            items['adapter']['prefix'] = self.adapter.get('prefix')
            items['adapter']['separator'] = self.adapter.get('separator')
            item_arg = make_argument(items, 'name')
            values = []
            for val in self.value:
                item_arg.value = val
                values += item_arg.cli()

            return values
ARGUMENT_TYPES['array'] = ArrayArgument


class ObjectArgument(InputArgument):

    def __init__(self, name, schema):
        self.schemas.append(super(ObjectArgument, self).__init__(name, schema))
        if schema.get('properties'):
            pass
        elif schema.get('oneOf'):
            pass
        else:
            raise RuntimeError('Invalid object type')

    def cli(self):
        [make_argument(input_spec, input_name)
         for input_name, input_spec
         in self.schema['properties'].iteritems()]
        pass
ARGUMENT_TYPES['object'] = ObjectArgument


class Adapter(object):
    def __init__(self, tool):
        self.tool = tool
        self.args = []

        if tool['adapter'].get('args'):
            self.args += [make_argument(arg) for arg
                          in tool['adapter']['args']]

        self.args += [make_argument(input_spec, input_name)
                      for input_name, input_spec
                      in tool['inputs']['properties'].iteritems()]
        sorted(self.args, key=lambda a: a.weight)

    def cli(self, job):
        for req in self.tool['inputs']['required']:
            if req not in job['inputs']:
                raise RuntimeError("Required input not provided: " + req)

        cli = list(self.tool['adapter']['baseCmd'])
        for arg in self.args:
            arg.bind(job)
            if arg.value is not None:
                cli += arg.cli()
        return cli


def gen_cli(tool, job):
    a = Adapter(tool)
    return a.cli(job)


# # Here be tests
from nose.tools import eq_

TOOL_STUB = {
    'inputs': {
        'type': 'object',
        'required': [],
        'properties': {},
        'adapter': {
            'baseCmd': [],
            'args': []
        }
    }
}


def test_simple_argument():
    arg = make_argument({'order': 1, 'value': 5})
    eq_(arg.cli(), ['5'])


def test_ref_argument():
    arg = make_argument({'order': 1, 'valueFrom': '#ref'})
    arg.bind({'ref': 'value'})
    eq_(arg.cli(), ['value'])


def test_argument_separator():
    arg = make_argument({'order': 1, 'value': 'str', 'prefix': '-x'})
    eq_(arg.cli(), ['-x', 'str'])

    arg = make_argument({'value': 'str', 'prefix': '-x', 'separator': '='})
    eq_(arg.cli(), ['-x=str'])


def test_argument_transform():
    arg = make_argument({'value': 'str.ext',
                         'transform': 'strip_ext'})
    eq_(arg.cli(), ['str'])


def test_file_argument():
    arg = make_argument({'type': 'file'}, 'ref')
    arg.bind({'inputs': {'ref': {'path': 'a/path'}}})
    eq_(arg.cli(), ['a/path'])


def test_list_argument():
    arg = make_argument({'type': 'array',
                         'items': {'type': 'number'},
                         'adapter': {'prefix': '-x'}
                         }, 'ref')
    arg.bind({'inputs': {'ref': [1, 2, 3]}})
    eq_(arg.cli(), ['-x', '1', '-x', '2', '-x', '3'])

    arg = make_argument({'type': 'array',
                         'items': {'type': 'number'},
                         'adapter': {'prefix': '-x', 'item_separator': ','}
                         }, 'ref')
    arg.bind({'inputs': {'ref': [1, 2, 3]}})
    eq_(arg.cli(), ['-x', '1,2,3'])


def test_list_argument_file_transform():
    arg = make_argument({'adapter': {'prefix': '-x'},
                         'type': 'array',
                         'items': {'type': 'file',
                                   'adapter': {'transform': 'strip_ext'}}
                         }, 'value')
    arg.bind({'inputs': {'value': [{'path': 'a/b.txt'}, {'path': 'c/d.txt'}]}})
    eq_(arg.cli(), ['-x', 'a/b', '-x', 'c/d'])


def test_bwa_mem():
    path = os.path.join(os.path.dirname(__file__), '../examples/bwa-mem.yml')
    doc = from_url(path)
    tool, job = doc['tool'], doc['job']
    print(gen_cli(tool, job))


def test_tmap_mapall():
    path = os.path.join(os.path.dirname(__file__), '../examples/tmap.yml')
    doc = from_url(path)
    tool, job = doc['mapall'], doc['exampleJob']
    print(gen_cli(tool, job))


# test_bwa_mem()
# test_tmap_mapall()
